# 11.2 RTTI语法


Java用`Class`对象实现自己的RTTI功能——即便我们要做的只是象转换那样的一些工作。`Class`类也提供了其他大量方式，以方便我们使用RTTI。

首先必须获得指向适当`Class`对象的的一个引用。就象前例演示的那样，一个办法是用一个字符串以及`Class.forName()`方法。这是非常方便的，因为不需要那种类型的一个对象来获取`Class`引用。然而，对于自己感兴趣的类型，如果已有了它的一个对象，那么为了取得`Class`引用，可调用属于`Object`根类一部分的一个方法：`getClass()`。它的作用是返回一个特定的`Class`引用，用来表示对象的实际类型。`Class`提供了几个有趣且较为有用的方法，从下例即可看出：

```
//: ToyTest.java
// Testing class Class

interface HasBatteries {}
interface Waterproof {}
interface ShootsThings {}
class Toy {
  // Comment out the following default
  // constructor to see
  // NoSuchMethodError from (*1*)
  Toy() {}
  Toy(int i) {}
}

class FancyToy extends Toy
    implements HasBatteries,
      Waterproof, ShootsThings {
  FancyToy() { super(1); }
}

public class ToyTest {
  public static void main(String[] args) {
    Class c = null;
    try {
      c = Class.forName("FancyToy");
    } catch(ClassNotFoundException e) {}
    printInfo(c);
    Class[] faces = c.getInterfaces();
    for(int i = 0; i < faces.length; i++)
      printInfo(faces[i]);
    Class cy = c.getSuperclass();
    Object o = null;
    try {
      // Requires default constructor:
      o = cy.newInstance(); // (*1*)
    } catch(InstantiationException e) {}
      catch(IllegalAccessException e) {}
    printInfo(o.getClass());
  }
  static void printInfo(Class cc) {
    System.out.println(
      "Class name: " + cc.getName() +
      " is interface? [" +
      cc.isInterface() + "]");
  }
} ///:~
```

从中可以看出，`class FancyToy`相当复杂，因为它从`Toy`中继承，并实现了`HasBatteries`，`Waterproof`以及`ShootsThings`的接口。在`main()`中创建了一个`Class`引用，并用位于相应`try`块内的`forName()`初始化成`FancyToy`。

`Class.getInterfaces`方法会返回`Class`对象的一个数组，用于表示包含在`Class`对象内的接口。

若有一个`Class`对象，也可以用`getSuperclass()`查询该对象的直接基类是什么。当然，这种做会返回一个`Class`引用，可用它作进一步的查询。这意味着在运行期的时候，完全有机会调查到对象的完整层次结构。

若从表面看，`Class`的`newInstance()`方法似乎是克隆（`clone()`）一个对象的另一种手段。但两者是有区别的。利用`newInstance()`，我们可在没有现成对象供“克隆”的情况下新建一个对象。就象上面的程序演示的那样，当时没有`Toy`对象，只有`cy`——即`y`的`Class`对象的一个引用。利用它可以实现“虚拟构造器”。换言之，我们表达：“尽管我不知道你的准确类型是什么，但请你无论如何都正确地创建自己。”在上述例子中，`cy`只是一个`Class`引用，编译期间并不知道进一步的类型信息。一旦新建了一个实例后，可以得到`Object`引用。但那个引用指向一个`Toy`对象。当然，如果要将除`Object`能够接收的其他任何消息发出去，首先必须进行一些调查研究，再进行转换。除此以外，用`newInstance()`创建的类必须有一个默认构造器。没有办法用`newInstance()`创建拥有非默认构造器的对象，所以在Java 1.0中可能存在一些限制。然而，Java 1.1的“反射”API（下一节讨论）却允许我们动态地使用类里的任何构造器。

程序中的最后一个方法是`printInfo()`，它取得一个`Class`引用，通过`getName()`获得它的名字，并用`interface()`调查它是不是一个接口。

该程序的输出如下：

```
Class name: FancyToy is interface? [false]
Class name: HasBatteries is interface? [true]
Class name: Waterproof is interface? [true]
Class name: ShootsThings is interface? [true]
Class name: Toy is interface? [false]
```

所以利用`Class`对象，我们几乎能将一个对象的祖宗十八代都调查出来。
